/**
 * Copyright (c) 2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl;

import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;
import org.ow2.petals.probes.api.probes.GaugeProbe;
import org.ow2.petals.probes.api.sensor.GaugeSensor;
import org.ow2.petals.probes.api.sensor.detector.GaugeDefectDetector;

/**
 * <p>
 * A gauge probe to probe instant value on demand, and storing the max reached
 * value (that can be greater than the max of instant values) since its last
 * start.</ul>
 * </p>
 * <p>
 * <u><b>Lifecycle</b>:</u> The instant value can be acceded or picked only if
 * the probe is started. The max value can be acceded even if the probe is
 * initialized or stopped. The max value is reseted to the sensor initial value
 * when the probe starts.
 * </p>
 * 
 * @param <T>
 *            The type of measured values (internal value type).
 * @param <S>
 *            The type of value to return to client probe (external value type).
 * 
 * @author Christophe DENEUX (PetalsLink)
 * 
 */
public class GaugeProbeImpl<T extends Comparable<T>, S> extends AbstractProbe implements
        GaugeProbe<T, S> {

    /**
     * Max value of the probe
     */
    private T maxValue;

    /**
     * <p>
     * Synchronization object to access read/write the {@link #maxValue}.
     * </p>
     * <p>
     * A dedicated object is needed because instance of {@link #maxValue} will
     * be re-assigned when a new max value is detected. So the lock will not
     * work correctly.
     * </p>
     */
    private final Object lockMaxValue = new Object();

    /**
     * The sensor to access instant value
     */
    protected final GaugeSensor<T, S> sensor;

    /**
     * The defect detector
     */
    protected final GaugeDefectDetector<T> defectDetector;

    /**
     * 
     * @param sensor
     *            The sensor. Cannot be <code>null</code>.
     * @param defectDetector
     *            The defect detector. Can be <code>null</code>.
     */
    public GaugeProbeImpl(final GaugeSensor<T, S> sensor,
            final GaugeDefectDetector<T> defectDetector)
            throws IllegalArgumentException {
        if (sensor == null) {
            throw new IllegalArgumentException("The sensor can not be null");
        }
        this.sensor = sensor;
        this.maxValue = sensor.getInitialValue();
        this.defectDetector = defectDetector;
    }

    /**
     * <p>
     * Realizes a measure and compare the value to the max one to update it.
     * </p>
     * <p>
     * This method <b>MUST</b> be invoked by the application internals when the
     * resource associated to the probe changes.
     * </p>
     * <p>
     * This method is thread-safe
     * </p>
     * 
     * @throws ProbeNotStartedException
     *             The probe is not started.
     */
    public void pick() throws ProbeNotStartedException {
        this.localInstantValue();
    }

    /**
     * <p>
     * Realizes a measure and compare the value to the max one to update it.
     * </p>
     * <p>
     * This method is thread-safe
     * </p>
     * 
     * @throws ProbeNotStartedException
     *             The probe is not started.
     */
    private T localInstantValue() throws ProbeNotStartedException {
        if (this.isStarted) {
            final T instantValue = this.sensor.getInstantValue();
            synchronized (this.lockMaxValue) {
                if (instantValue.compareTo(this.maxValue) > 0) {
                    this.maxValue = instantValue;
                }
                ;
            }
            if (this.defectDetector != null) {
                this.defectDetector.detect(instantValue);
            }
            return instantValue;
        } else {
            throw new ProbeNotStartedException();
        }
    }

    /**
     * <p>
     * Accessor to the sensor instant value.
     * </p>
     * <p>
     * This method <b>MUST</b> be invoked by monitoring clients to get the probe
     * instant value.
     * </p>
     * <p>
     * This method is thread-safe
     * </p>
     * 
     * @return The sensor instant value.
     * @throws ProbeNotStartedException
     *             The probe is not started.
     */
    public S getInstantValue() throws ProbeNotStartedException {
        final T instantValue = this.localInstantValue();
        return this.sensor.toExternalValue(instantValue);
    }

    /**
     * <p>
     * Accessor to the max value.
     * </p>
     * <p>
     * This method is thread-safe
     * </p>
     * 
     * @return The probe max value.
     * @throws ProbeNotInitializedException
     *             The probe is not initialized.
     */
    public S getMaxValue() throws ProbeNotInitializedException {
        if (this.isInitialized) {
            synchronized (this.lockMaxValue) {
                return this.sensor.toExternalValue(this.maxValue);
            }
        } else {
            throw new ProbeNotInitializedException();
        }
    }

    @Override
    public final void doInit() throws ProbeInitializationException {
        // NOP
    }

    @Override
    public final void doStart() throws ProbeStartupException {
        // Reset the max value
        synchronized (this.lockMaxValue) {
            this.maxValue = this.sensor.getInitialValue();
        }
    }

    @Override
    public final void doStop() throws ProbeStopException {
        // NOP
    }

    @Override
    public final void doShutdown() throws ProbeShutdownException {
        // NOP
    }
}
