/**
 * Copyright (c) 2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import org.ow2.petals.probes.api.enums.ExecutionStatus;
import org.ow2.petals.probes.api.exceptions.NoResponseTimeException;
import org.ow2.petals.probes.api.exceptions.ResponseTimeCollectionStoppedException;
import org.ow2.petals.probes.api.exceptions.StartDateItemLostException;
import org.ow2.petals.probes.api.exceptions.StartDateItemUnknownException;
import org.ow2.petals.probes.api.probes.ResponseTimeRelativeValues;
import org.ow2.petals.probes.api.probes.StartDateItem;

/**
 * <p>
 * Response time sample collecting response times during its sampling duration.
 * When collection ends, relative response time values are computed.
 * </p>
 * <p>
 * Relative response time values are defined by
 * {@link ResponseTimeRelativeValues}
 * </p>
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ResponseTimesSample {

    /**
     * <p>
     * List of start date of pending executions. When a execution ends, the end
     * time is available and the response is moved to {@link #responseTimes}.
     * </p>
     */
    private final List<StartDateItem> startDates;

    /**
     * <p>
     * List of response times (of ended executions) of the current sample. This
     * list is filled when the sample runs. This list is clear when the sample
     * ends. The content of this list is used to update the max response time
     * values.
     * </p>
     * <p>
     * <u>Note:</u> {@link LinkedList} is used in definition instead of
     * {@link List} because of performance reasons, access to a linked list
     * elements is faster with the {@link LinkedList} API than using the
     * {@link List} API.
     * </p>
     */
    private final ConcurrentMap<ExecutionStatus, LinkedList<Long>> responseTimes = new ConcurrentHashMap<ExecutionStatus, LinkedList<Long>>();

    /**
     * List of response time values (max, avg, min, percentile, ...) of the
     * current sample. It is computed from response times available into
     * {@link #responseTimes} and {@link #startDates} when the sample is
     * stopped.
     */
    private final Map<ExecutionStatus, ResponseTimeRelativeValues> responseTimeValues = new HashMap<ExecutionStatus, ResponseTimeRelativeValues>();

    /**
     * <p>
     * Flag to control the sample lifecycle. When the sample ends, no more
     * response times is collected and response time values (max, avg, min,
     * percentiles, ...) are computed.
     * </p>
     */
    private boolean isEchantillonRunning = true;

    /**
     * Lock object to synchronize the collection of response times
     */
    private final Object stopCollectionLock = new Object();

    /**
     * <p>
     * Creates a new sample from scratch. It <b>MUST</b> be the constructor to
     * use to create the <b>FIRST</b> sample.
     * </p>
     */
    public ResponseTimesSample() {
        this.startDates = new LinkedList<StartDateItem>();
    }

    /**
     * <p>
     * Creates a new sample from the current one:
     * <ul>
     * <li>the start dates are transfered,</li>
     * <li>the response times are not transfered,</li>
     * <li>the sample status is reset.</li>
     * </ul>
     * </p>
     */
    public ResponseTimesSample(final ResponseTimesSample currentRelativeResponseTimes) {
        this.startDates = currentRelativeResponseTimes.startDates;
    }

    /**
     * Add a new start date of a pending execution.
     * 
     * @param startTime
     *            The start time (timestamp)
     * @return A {@link StartDateItem} created for this new execution.
     * @throws ResponseTimeCollectionStoppedException
     *             The sample collecting response times is stopped.
     * @throws StartDateItemLostException
     *             Unable to add the new start date to the list of start dates.
     *             This exception should not occur.
     */
    public StartDateItem addStartDate(final long startTime)
            throws ResponseTimeCollectionStoppedException, StartDateItemLostException {
        synchronized (this.stopCollectionLock) {
            if (this.isEchantillonRunning) {
                final StartDateItem startDateItem = new StartDateItem(startTime);
                synchronized (this.startDates) {
                    if (this.startDates.add(startDateItem)) {
                        return startDateItem;
                    } else {
                        throw new StartDateItemLostException();
                    }
                }
            } else {
                throw new ResponseTimeCollectionStoppedException();
            }
        }
    }

    /**
     * The execution associated to <code>startDateItem</code> ends.
     * 
     * @param startDateItem
     *            The {@link StartDateItem} created by
     *            {@link #addStartDate(long)} for the current execution.
     * @param endTime
     *            The time where the execution ends (timestamp)
     * @param executionStatus
     *            The execution status with which the execution ends
     * @throws ResponseTimeCollectionStoppedException
     *             The sample collecting response times is stopped. This
     *             execution should not occur.
     * @throws StartDateItemUnknownException
     *             <code>startDateItem</code> not found in the list of start
     *             dates. This exception should not occur.
     */
    public void endStartDate(final StartDateItem startDateItem, final long endTime,
            final ExecutionStatus executionStatus) throws ResponseTimeCollectionStoppedException,
            StartDateItemUnknownException {
        synchronized (this.stopCollectionLock) {
            if (this.isEchantillonRunning) {
                synchronized (this.startDates) {
                    if (!this.startDates.remove(startDateItem)) {
                        throw new StartDateItemUnknownException(startDateItem, executionStatus);
                    }
                }

                // Add the response time to its list
                final Long responseTime = endTime - startDateItem.startDate;
                final LinkedList<Long> emptyResponseTimeList = new LinkedList<Long>();
                final List<Long> responseTimeList = this.responseTimes.putIfAbsent(executionStatus,
                        emptyResponseTimeList);
                if (responseTimeList != null) {
                    responseTimeList.add(responseTime);
                } else {
                    emptyResponseTimeList.add(responseTime);
                }
            } else {
                throw new ResponseTimeCollectionStoppedException();
            }
        }
    }

    /**
     * Stops the sample to collect response times, and compute response time
     * values.
     */
    public void stopCollecting() {
        synchronized (this.stopCollectionLock) {

            // Stop collecting
            this.isEchantillonRunning = false;

            // Timestamp of the collecting stop (needed to compute elapsed time
            // of pending execution)
            final long stopTimestamp = System.currentTimeMillis();

            // Compute response time values
            final Iterator<Map.Entry<ExecutionStatus, LinkedList<Long>>> itResponseTimes = this.responseTimes
                    .entrySet().iterator();
            while (itResponseTimes.hasNext()) {
                final Map.Entry<ExecutionStatus, LinkedList<Long>> responseTimesEntry = itResponseTimes
                        .next();
                final LinkedList<Long> responseTimes = responseTimesEntry.getValue();
                if (responseTimes.size() != 0) {
                    try {
                        this.responseTimeValues.put(responseTimesEntry.getKey(),
                                new ResponseTimeRelativeValues(responseTimesEntry.getValue()));
                    } catch (final NoResponseTimeException e) {
                        // This exception should not occur because of the
                        // previous check about number of response times
                    }
                } else {
                    // No response time collected for the key.
                }
            }

            // Add the pending execution to the response time values.
            final LinkedList<Long> pendingStartDates = new LinkedList<Long>();
            synchronized (this.startDates) {
                final Iterator<StartDateItem> itStartDates = this.startDates.iterator();
                while (itStartDates.hasNext()) {
                    final StartDateItem startDate = itStartDates.next();
                    final long responseTime = stopTimestamp - startDate.startDate;
                    pendingStartDates.add(responseTime);
                }
            }
            try {
                this.responseTimeValues.put(ExecutionStatus.PENDING,
                        new ResponseTimeRelativeValues(pendingStartDates));
            } catch (final NoResponseTimeException e) {
                // This exception should not occur because of the
                // previous check about number of response times
            }
        }
    }

    /**
     * @return The relative response time values
     */
    public Map<ExecutionStatus, ResponseTimeRelativeValues> getResponseTimeValues() {
        synchronized (this.stopCollectionLock) {
            return this.responseTimeValues;
        }
    }
}
