/**
 * Copyright (c) 2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl.macro;

import java.util.concurrent.ThreadPoolExecutor;

import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;
import org.ow2.petals.probes.api.probes.GaugeProbe;
import org.ow2.petals.probes.api.probes.macro.ThreadPoolProbe;
import org.ow2.petals.probes.api.sensor.detector.DefectCreator;
import org.ow2.petals.probes.impl.GaugeProbeImpl;
import org.ow2.petals.probes.impl.macro.detector.ThreadPoolAllocatedThreadsDefectDetector;
import org.ow2.petals.probes.impl.macro.sensor.ThreadPoolAllocatedThreadsGaugeSensor;
import org.ow2.petals.probes.impl.macro.sensor.ThreadPoolIdleThreadsGaugeSensor;
import org.ow2.petals.probes.impl.macro.sensor.ThreadPoolQueuedRequestsGaugeSensor;

/**
 * <p>
 * A macro containing all probes needed to monitor a thread pool
 * </p>
 * <p>
 * <u><b>Lifecycle</b>:</u> The instant value can be acceded or picked only if
 * the probe is started. The max value can be acceded even if the probe is
 * initialized or stopped. The max value is reseted to the sensor initial value
 * when the probe starts.
 * </p>
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ThreadPoolProbeImpl implements ThreadPoolProbe {

    /**
     * The probe counting allocated threads in the thread pool.
     */
    private GaugeProbe<Long, Long> probeThreadPoolAllocatedThreads;

    /**
     * The probe counting idle threads in the thread pool.
     */
    private GaugeProbe<Long, Long> probeThreadPoolIdleThreads;

    /**
     * The probe counting requests in the queue of the thread pool.
     */
    private GaugeProbe<Long, Long> probeThreadPoolQueuedRequests;

    /**
     * Sensor attached to the allocated threads of the thread pool.
     */
    private final ThreadPoolAllocatedThreadsGaugeSensor threadPoolAllocatedThreadsGaugeSensor;

    /**
     * Defect detector attached to the allocated threads of the thread pool.
     */
    private final ThreadPoolAllocatedThreadsDefectDetector threadPoolAllocatedThreadsDefectDetector;

    /**
     * Sensor attached to the idle threads of the thread pool.
     */
    private final ThreadPoolIdleThreadsGaugeSensor threadPoolIdleThreadsGaugeSensor;

    /**
     * Sensor attached to the queued requests of the thread pool.
     */
    private final ThreadPoolQueuedRequestsGaugeSensor threadPoolQueuedRequestsGaugeSensor;

    /**
     * The probed thread pool
     */
    private ThreadPoolExecutor threadPool = null;

    /**
     * @param threadPoolAllocatedThreadsDefectCreator
     *            Defect creator used to create defects about exhaustion of the
     *            thread pool
     */
    public ThreadPoolProbeImpl(final DefectCreator threadPoolAllocatedThreadsDefectCreator) {

        // Create the probe about allocated threads
        this.threadPoolAllocatedThreadsGaugeSensor = new ThreadPoolAllocatedThreadsGaugeSensor();
        this.threadPoolAllocatedThreadsDefectDetector = new ThreadPoolAllocatedThreadsDefectDetector(
                threadPoolAllocatedThreadsDefectCreator);
        this.probeThreadPoolAllocatedThreads = new GaugeProbeImpl<Long, Long>(
                this.threadPoolAllocatedThreadsGaugeSensor,
                this.threadPoolAllocatedThreadsDefectDetector);

        // Create the probe about idle threads
        this.threadPoolIdleThreadsGaugeSensor = new ThreadPoolIdleThreadsGaugeSensor();
        this.probeThreadPoolIdleThreads = new GaugeProbeImpl<Long, Long>(
                this.threadPoolIdleThreadsGaugeSensor, null);

        // Create the probe about enqueued tasks
        this.threadPoolQueuedRequestsGaugeSensor = new ThreadPoolQueuedRequestsGaugeSensor();
        this.probeThreadPoolQueuedRequests = new GaugeProbeImpl<Long, Long>(
                this.threadPoolQueuedRequestsGaugeSensor, null);

    }

    @Override
    public void setThreadPool(final ThreadPoolExecutor threadPool) {
        this.threadPool = threadPool;
        this.threadPoolAllocatedThreadsDefectDetector.setThreadPool(this.threadPool);
        this.threadPoolAllocatedThreadsGaugeSensor.setThreadPool(this.threadPool);
        this.threadPoolIdleThreadsGaugeSensor.setThreadPool(this.threadPool);
        this.threadPoolQueuedRequestsGaugeSensor.setThreadPool(this.threadPool);
    }

    @Override
    public long getThreadPoolMaxSize() {
        if (this.threadPool == null) {
            return 0L;
        } else {
            return this.threadPool.getMaximumPoolSize();
        }
    }

    @Override
    public long getThreadPoolMinSize() {
        if (this.threadPool == null) {
            return 0L;
        } else {
            return this.threadPool.getCorePoolSize();
        }
    }

    @Override
    public long getThreadPoolAllocatedThreadsMax() throws ProbeNotInitializedException {
        return this.probeThreadPoolAllocatedThreads.getMaxValue();
    }

    @Override
    public long getThreadPoolAllocatedThreadsCurrent() throws ProbeNotStartedException {
        return this.probeThreadPoolAllocatedThreads.getInstantValue();
    }

    @Override
    public long getThreadPoolIdleThreadsMax() throws ProbeNotInitializedException {
        return this.probeThreadPoolIdleThreads.getMaxValue();
    }

    @Override
    public long getThreadPoolIdleThreadsCurrent() throws ProbeNotStartedException {
        return this.probeThreadPoolIdleThreads.getInstantValue();
    }

    @Override
    public long getThreadPoolQueuedRequestsMax() throws ProbeNotInitializedException {
        return this.probeThreadPoolQueuedRequests.getMaxValue();
    }

    @Override
    public long getThreadPoolQueuedRequestsCurrent() throws ProbeNotStartedException {
        return this.probeThreadPoolQueuedRequests.getInstantValue();
    }

    @Override
    public void pick() throws ProbeNotStartedException {
        this.probeThreadPoolAllocatedThreads.pick();
        this.probeThreadPoolIdleThreads.pick();
        this.probeThreadPoolQueuedRequests.pick();
    }

    @Override
    public void init() throws ProbeInitializedException, ProbeStartedException,
            ProbeInitializationException {
        this.probeThreadPoolAllocatedThreads.init();
        this.probeThreadPoolIdleThreads.init();
        this.probeThreadPoolQueuedRequests.init();
    }

    @Override
    public void start() throws ProbeNotInitializedException, ProbeStartedException,
            ProbeStartupException {
        this.probeThreadPoolAllocatedThreads.start();
        this.probeThreadPoolIdleThreads.start();
        this.probeThreadPoolQueuedRequests.start();
    }

    @Override
    public void stop() throws ProbeNotInitializedException, ProbeNotStartedException,
            ProbeStopException {
        this.probeThreadPoolAllocatedThreads.stop();
        this.probeThreadPoolIdleThreads.stop();
        this.probeThreadPoolQueuedRequests.stop();
    }

    @Override
    public void shutdown() throws ProbeShutdownException, ProbeStartedException,
            ProbeNotInitializedException {
        this.probeThreadPoolAllocatedThreads.shutdown();
        this.probeThreadPoolIdleThreads.shutdown();
        this.probeThreadPoolQueuedRequests.shutdown();
    }

}
