/**
 * Copyright (c) 2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl;

import junit.framework.TestCase;

import org.ow2.petals.probes.api.exceptions.ProbeException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;

/**
 * <p>
 * Unit tests of {@link AbstractProbe} to verify the lifecycle management.
 * </p>
 * 
 * @author Christophe DENEUX (PetalsLink)
 * 
 */
public class AbstractProbeTestCase extends TestCase {

    private class TestProbeSpecificDoMethods extends AbstractProbe {

        protected volatile short specificInitializationDoneCount = 0;

        protected volatile short specificStartupDoneCount = 0;

        protected volatile short specificStopDoneCount = 0;

        protected volatile short specificShutdownDoneCount = 0;

        /**
         * Flag to drive the processing of overridden method 'doInit':
         * exceptions must be thrown or not.
         */
        private final boolean isWithErrorForInit;

        /**
         * Flag to drive the processing of overridden method 'doStart':
         * exceptions must be thrown or not.
         */
        private final boolean isWithErrorForStart;

        /**
         * Flag to drive the processing of overridden method 'doStop':
         * exceptions must be thrown or not.
         */
        private final boolean isWithErrorForStop;

        /**
         * Flag to drive the processing of overridden method 'doShutdown':
         * exceptions must be thrown or not.
         */
        private final boolean isWithErrorForShutdown;

        /**
         * Creates a new test probe.
         * 
         * @param isWithError
         *            Enable or disable exception throwing by overriden methods.
         */
        public TestProbeSpecificDoMethods(final boolean isWithErrorForInit,
                final boolean isWithErrorForStart, final boolean isWithErrorForStop,
                final boolean isWithErrorForShutdown) {
            this.isWithErrorForInit = isWithErrorForInit;
            this.isWithErrorForStart = isWithErrorForStart;
            this.isWithErrorForStop = isWithErrorForStop;
            this.isWithErrorForShutdown = isWithErrorForShutdown;
        }

        @Override
        public void doInit() throws ProbeInitializationException {
            if (this.isWithErrorForInit) {
                throw new ProbeInitializationException("Error during probe initialization.");
            } else {
                this.specificInitializationDoneCount++;
            }
        }

        @Override
        public void doStart() throws ProbeStartupException {
            if (this.isWithErrorForStart) {
                throw new ProbeStartupException("Error during probe startup.");
            } else {
                this.specificStartupDoneCount++;
            }
        }

        @Override
        public void doStop() throws ProbeStopException {
            if (this.isWithErrorForStop) {
                throw new ProbeStopException("Error during probe stop.");
            } else {
                this.specificStopDoneCount++;
            }
        }

        @Override
        public void doShutdown() throws ProbeShutdownException {
            if (this.isWithErrorForShutdown) {
                throw new ProbeShutdownException("Error during probe shutdown.");
            } else {
                this.specificShutdownDoneCount++;
            }
        }
    }

    /*----------------------*/
    /* Lifecycle test cases */
    /*----------------------*/

    /**
     * <p>
     * Tests the initialization of a just allocated probe.
     * </p>
     * <p>
     * Expected results: no exception is thrown.
     * </p>
     */
    public void testLifecycle_001() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization
        try {
            testProbe.init();
            TestCase.assertTrue("The probe is not initialized", testProbe.isInitialized);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' was not invoked once.", 1,
                    testProbe.specificInitializationDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the startup of a initialized probe.
     * </p>
     * <p>
     * Expected results: no exception is thrown.
     * </p>
     */
    public void testLifecycle_002() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization and startup
        try {
            testProbe.init();
            TestCase.assertTrue("The probe is not initialized", testProbe.isInitialized);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' was not invoked once.", 1,
                    testProbe.specificInitializationDoneCount);

            testProbe.start();
            TestCase.assertTrue("The probe is not started", testProbe.isStarted);
            TestCase.assertEquals("The specific startup method 'doStart' was not invoked once.", 1,
                    testProbe.specificStartupDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the stop of a started probe.
     * </p>
     * <p>
     * Expected results: no exception is thrown.
     * </p>
     */
    public void testLifecycle_003() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization, startup and stop
        try {
            testProbe.init();
            TestCase.assertTrue("The probe is not initialized", testProbe.isInitialized);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' was not invoked once.", 1,
                    testProbe.specificInitializationDoneCount);

            testProbe.start();
            TestCase.assertTrue("The probe is not started", testProbe.isStarted);
            TestCase.assertEquals("The specific startup method 'doStart' was not invoked once.", 1,
                    testProbe.specificStartupDoneCount);

            testProbe.stop();
            TestCase.assertFalse("The probe is started", testProbe.isStarted);
            TestCase.assertEquals("The specific stop method 'doStopt' was not invoked once.", 1,
                    testProbe.specificStopDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the startup of a stopped probe.
     * </p>
     * <p>
     * Expected results: no exception is thrown.
     * </p>
     */
    public void testLifecycle_004() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization, startup, stop and startup
        try {
            testProbe.init();
            TestCase.assertTrue("The probe is not initialized", testProbe.isInitialized);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' was not invoked once.", 1,
                    testProbe.specificInitializationDoneCount);

            testProbe.start();
            TestCase.assertTrue("The probe is not started", testProbe.isStarted);
            TestCase.assertEquals("The specific startup method 'doStart' was not invoked once.", 1,
                    testProbe.specificStartupDoneCount);

            testProbe.stop();
            TestCase.assertFalse("The probe is started", testProbe.isStarted);
            TestCase.assertEquals("The specific stop method 'doStopt' was not invoked once.", 1,
                    testProbe.specificStopDoneCount);

            testProbe.start();
            TestCase.assertTrue("The probe is not started", testProbe.isStarted);
            TestCase.assertEquals("The specific startup method 'doStart' was not invoked twice.",
                    2, testProbe.specificStartupDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the shutdown of a stopped probe.
     * </p>
     * <p>
     * Expected results: no exception is thrown.
     * </p>
     */
    public void testLifecycle_005() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization, startup, stop and shutdown
        try {
            testProbe.init();
            TestCase.assertTrue("The probe is not initialized", testProbe.isInitialized);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' was not invoked once.", 1,
                    testProbe.specificInitializationDoneCount);

            testProbe.start();
            TestCase.assertTrue("The probe is not started", testProbe.isStarted);
            TestCase.assertEquals("The specific startup method 'doStart' was not invoked once.", 1,
                    testProbe.specificStartupDoneCount);

            testProbe.stop();
            TestCase.assertFalse("The probe is started", testProbe.isStarted);
            TestCase.assertEquals("The specific stop method 'doStopt' was not invoked once.", 1,
                    testProbe.specificStopDoneCount);

            testProbe.shutdown();
            TestCase.assertFalse("The probe is initialized", testProbe.isInitialized);
            TestCase.assertEquals(
                    "The specific shutdown method 'doShutdown' was not invoked once.", 1,
                    testProbe.specificShutdownDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the initialization of a shutdown probe.
     * </p>
     * <p>
     * Expected result: no exception is thrown.
     * </p>
     */
    public void testLifecycle_006() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization, startup, stop, shutdown and initialization
        try {
            testProbe.init();
            TestCase.assertTrue("The probe is not initialized", testProbe.isInitialized);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' was not invoked once.", 1,
                    testProbe.specificInitializationDoneCount);

            testProbe.start();
            TestCase.assertTrue("The probe is not started", testProbe.isStarted);
            TestCase.assertEquals("The specific startup method 'doStart' was not invoked once.", 1,
                    testProbe.specificStartupDoneCount);

            testProbe.stop();
            TestCase.assertFalse("The probe is started", testProbe.isStarted);
            TestCase.assertEquals("The specific stop method 'doStopt' was not invoked once.", 1,
                    testProbe.specificStopDoneCount);

            testProbe.shutdown();
            TestCase.assertFalse("The probe is initialized", testProbe.isInitialized);
            TestCase.assertEquals(
                    "The specific shutdown method 'doShutdown' was not invoked once.", 1,
                    testProbe.specificShutdownDoneCount);

            testProbe.init();
            TestCase.assertTrue("The probe is not initialized", testProbe.isInitialized);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' was not invoked twice.", 2,
                    testProbe.specificInitializationDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the shutdown of a initialized probe.
     * </p>
     * <p>
     * Expected result: no exception is thrown.
     * </p>
     */
    public void testLifecycle_007() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization and shutdown
        try {
            testProbe.init();
            TestCase.assertTrue("The probe is not initialized", testProbe.isInitialized);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' was not invoked once.", 1,
                    testProbe.specificInitializationDoneCount);

            testProbe.shutdown();
            TestCase.assertFalse("The probe is initialized", testProbe.isInitialized);
            TestCase.assertEquals(
                    "The specific shutdown method 'doShutdown' was not invoked once.", 1,
                    testProbe.specificShutdownDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests a specific initialization.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>the specific initialization is done through a call to the method
     * {@link AbstractProbe#doInit()}.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_008() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization
        try {
            testProbe.init();
            TestCase.assertEquals(
                    "The method 'doInit' was not invoke during the probe initialization.", 1,
                    testProbe.specificInitializationDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests a specific startup.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>the specific startup is done through a call to the method
     * {@link AbstractProbe#doStart()}.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_009() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization
        try {
            testProbe.init();
            testProbe.start();
            TestCase.assertEquals("The method 'doStart' was not invoke during the probe startup.",
                    1, testProbe.specificStartupDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests a specific stop.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>the specific stop is done through a call to the method
     * {@link AbstractProbe#doStop()}.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_010() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization
        try {
            testProbe.init();
            testProbe.start();
            testProbe.stop();
            TestCase.assertEquals("The method 'doStop' was not invoke during the probe stop.", 1,
                    testProbe.specificStopDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests a specific shutdown.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>the specific shutdown is done through a call to the method
     * {@link AbstractProbe#doShutdown()}.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_011() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization
        try {
            testProbe.init();
            testProbe.start();
            testProbe.stop();
            testProbe.shutdown();
            TestCase.assertEquals(
                    "The method 'doShutdown' was not invoke during the probe shutdown.", 1,
                    testProbe.specificShutdownDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /*----------------------------*/
    /* Lifecycle error test cases */
    /*----------------------------*/

    /**
     * <p>
     * Tests the initialization of a initialized probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeInitializedException} is thrown,</li>
     * <li>the specific method 'doInit' is not invoked,</li>
     * <li>the probe status does not change, always 'initialized'.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_100() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // First initialization
        try {
            testProbe.init();

            TestCase.assertTrue("The probe is not initialized", testProbe.isInitialized);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' was not invoked once.", 1,
                    testProbe.specificInitializationDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }

        // 2nd initialization
        try {
            testProbe.init();
            TestCase.fail("The 2nd initialization of a probe should fail");
        } catch (final ProbeInitializedException e) {
            // NOP, it's the expected exception
            TestCase.assertTrue("The probe is no more initialized", testProbe.isInitialized);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' should not be invoked on an initialization of an initialized probe.",
                    1, testProbe.specificInitializationDoneCount);
        } catch (final ProbeInitializationException e) {
            TestCase.fail(e.getMessage());
        } catch (final ProbeStartedException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the initialization of a started probe.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the exception {@link ProbeStartedException} is thrown,</li>
     * <li>the specific method 'doInit' is not invoked,</li>
     * <li>the specific method 'doStart' is not invoked,</li>
     * <li>the probe status does not change, always 'started'.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_101() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization and startup, no exception is expected.
        try {
            testProbe.init();
            testProbe.start();

            TestCase.assertTrue("The probe is not started", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' was not invoked once.", 1,
                    testProbe.specificInitializationDoneCount);
            TestCase.assertEquals("The specific startup method 'doStart' was not invoked once.", 1,
                    testProbe.specificStartupDoneCount);

        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }

        // 2nd initialization, ProbeStartedException is expected
        try {
            testProbe.init();
            TestCase.fail("The initialization of a started probe should fail with ProbeStartedException");
        } catch (final ProbeInitializedException e) {
            TestCase.fail(e.getMessage());
        } catch (final ProbeInitializationException e) {
            TestCase.fail(e.getMessage());
        } catch (final ProbeStartedException e) {
            // NOP, its the expected exception
            TestCase.assertTrue("The probe is no more started", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' should not be invoked on an initialization of a started probe.",
                    1, testProbe.specificInitializationDoneCount);
        }
    }

    /**
     * <p>
     * Tests the initialization of a stopped probe.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the exception {@link ProbeInitializedException} is thrown,</li>
     * <li>the specific method 'doInit' is not invoked,</li>
     * <li>the specific method 'doStart' is not invoked,</li>
     * <li>the specific method 'doStop' is not invoked,</li>
     * <li>the probe status does not change, always 'initialized' (ie. stopped).
     * </li>
     * </ul>
     * </p>
     */
    public void testLifecycle_102() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization and startup, no exception is expected.
        try {
            testProbe.init();
            testProbe.start();
            TestCase.assertTrue("The probe is not started", testProbe.isStarted);
            testProbe.stop();

            TestCase.assertFalse("The probe is not stopped", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' was not invoked once.", 1,
                    testProbe.specificInitializationDoneCount);
            TestCase.assertEquals("The specific startup method 'doStart' was not invoked once.", 1,
                    testProbe.specificStartupDoneCount);
            TestCase.assertEquals("The specific stop method 'doStop' was not invoked once.", 1,
                    testProbe.specificStartupDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }

        // 2nd initialization, ProbeInitiliazedException is expected
        try {
            testProbe.init();
            TestCase.fail("The initialization of a stopped probe should fail with ProbeStartedException");
        } catch (final ProbeInitializedException e) {
            // NOP, its the expected exception
            TestCase.assertTrue("The probe is no more initialized (ie. stopped)",
                    testProbe.isInitialized);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' should not be invoked on an initialization of a stopped probe.",
                    1, testProbe.specificInitializationDoneCount);
        } catch (final ProbeInitializationException e) {
            TestCase.fail(e.getMessage());
        } catch (final ProbeStartedException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the startup of a just allocated probe.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the exception {@link ProbeNotInitializedException} is thrown,</li>
     * <li>the specific method 'doStart' is not invoked,</li>
     * <li>the probe status does not change, always 'not initialized'.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_103() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Startup without initialization, ProbeNotInitiliazedException is
        // expected
        try {
            testProbe.start();
            TestCase.fail("The start of a not initialized probe should fail with ProbeNotInitializedException");
        } catch (final ProbeStartedException e) {
            TestCase.fail(e.getMessage());
        } catch (final ProbeNotInitializedException e) {
            // NOP, its the expected exception
            TestCase.assertFalse("The probe should not be initialized", testProbe.isInitialized);
            TestCase.assertFalse("The probe should not be started", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific startup method 'doStart' should not be invoked on a startup of a just allocated probe.",
                    0, testProbe.specificStartupDoneCount);
        } catch (final ProbeStartupException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the startup of a started probe.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the exception {@link ProbeStartedException} is thrown,</li>
     * <li>the specific method 'doStart' is not invoked,</li>
     * <li>the probe status does not change, always 'started'.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_104() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization and 1st startup without initialization, , no exception
        // is expected.
        try {
            testProbe.init();
            testProbe.start();

            TestCase.assertTrue("The probe is not started", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' was not invoked once.", 1,
                    testProbe.specificInitializationDoneCount);
            TestCase.assertEquals("The specific startup method 'doStart' was not invoked once.", 1,
                    testProbe.specificStartupDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }

        // 2nd startup, ProbeStartedException is expected
        try {
            testProbe.start();
            TestCase.fail("The start of a started probe should fail with ProbeStartedException");
        } catch (final ProbeStartedException e) {
            // NOP, its the expected exception
            TestCase.assertTrue("The probe is no more started", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific startup method 'doStart' should not be invoked on a startup of a started probe.",
                    1, testProbe.specificStartupDoneCount);
        } catch (final ProbeNotInitializedException e) {
            TestCase.fail(e.getMessage());
        } catch (final ProbeStartupException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the startup of a shutdown probe.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the exception {@link ProbeNotInitializedException} is thrown,</li>
     * <li>the specific method 'doStart' is not invoked,</li>
     * <li>the probe status does not change, always 'not initialized'.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_105() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization, startup, stop and shutdown, no exception is expected.
        try {
            testProbe.init();
            testProbe.start();
            testProbe.stop();
            testProbe.shutdown();

            TestCase.assertFalse("The probe is initialized", testProbe.isInitialized);
            TestCase.assertFalse("The probe is started", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' was not invoked once.", 1,
                    testProbe.specificInitializationDoneCount);
            TestCase.assertEquals("The specific startup method 'doStart' was not invoked once.", 1,
                    testProbe.specificStartupDoneCount);
            TestCase.assertEquals("The specific stop method 'doStop' was not invoked once.", 1,
                    testProbe.specificStopDoneCount);
            TestCase.assertEquals(
                    "The specific shutdown method 'doShutdown' was not invoked once.", 1,
                    testProbe.specificShutdownDoneCount);

        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }

        // 2nd startup, ProbeNotInitialized is expected
        try {
            testProbe.start();
            TestCase.fail("The start of a started probe should fail with ProbeStartedException");
        } catch (final ProbeStartedException e) {
            TestCase.fail(e.getMessage());
        } catch (final ProbeNotInitializedException e) {
            // NOP, its the expected exception
            TestCase.assertFalse("The probe is initialized", testProbe.isInitialized);
            TestCase.assertFalse("The probe is started", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific startup method 'doStart' should not be invoked on a startup of a shutdown probe.",
                    1, testProbe.specificStartupDoneCount);
        } catch (final ProbeStartupException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the stop of a just allocated probe.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the exception {@link ProbeNotInitializedException} is thrown,</li>
     * <li>the specific method 'doStart' is not invoked,</li>
     * <li>the probe status does not change, always 'not initialized'.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_106() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Stop without initialization, ProbeNotInitiliazedException is
        // expected
        try {
            testProbe.stop();
            TestCase.fail("The stop of a not initialized probe should fail with ProbeNotInitializedException");
        } catch (final ProbeStopException e) {
            TestCase.fail(e.getMessage());
        } catch (final ProbeNotInitializedException e) {
            // NOP, its the expected exception
            TestCase.assertFalse("The probe should not be initialized", testProbe.isInitialized);
            TestCase.assertFalse("The probe should not be started", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific stop method 'doStop' should not be invoked on a stop of a just allocated probe.",
                    0, testProbe.specificStopDoneCount);
        } catch (final ProbeNotStartedException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the stop of an initialized probe.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>the specific method 'doStop' is not invoked,</li>
     * <li>the probe status does not change, always 'initialized'.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_107() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization, no exception is expected.
        try {
            testProbe.init();

            TestCase.assertTrue("The probe is not initialized", testProbe.isInitialized);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' was not invoked once.", 1,
                    testProbe.specificInitializationDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }

        // Stop without startup, ProbeNotStartedException is
        // expected
        try {
            testProbe.stop();
            TestCase.fail("The stop of a not initialized probe should fail with ProbeNotInitializedException");
        } catch (final ProbeStopException e) {
            TestCase.fail(e.getMessage());
        } catch (final ProbeNotInitializedException e) {
            TestCase.fail(e.getMessage());
        } catch (final ProbeNotStartedException e) {
            // NOP, its the expected exception
            TestCase.assertTrue("The probe is no more initialized", testProbe.isInitialized);
            TestCase.assertFalse("The probe is started", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific stop method 'doStop' should not be invoked on a stop of an initialized probe.",
                    0, testProbe.specificStopDoneCount);
        }
    }

    /**
     * <p>
     * Tests the stop of a stopped probe.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>the specific method 'doStop' is not invoked,</li>
     * <li>the probe status does not change, always 'initialized'.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_108() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization, startup and stop, no exception is expected.
        try {
            testProbe.init();
            testProbe.start();
            testProbe.stop();

            TestCase.assertTrue("The probe is not initialized", testProbe.isInitialized);
            TestCase.assertFalse("The probe is started", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' was not invoked once.", 1,
                    testProbe.specificInitializationDoneCount);
            TestCase.assertEquals("The specific startup method 'doStart' was not invoked once.", 1,
                    testProbe.specificStartupDoneCount);
            TestCase.assertEquals("The specific stop method 'doStop' was not invoked once.", 1,
                    testProbe.specificStopDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }

        // Stop without startup, ProbeNotStartedException is
        // expected
        try {
            testProbe.stop();
            TestCase.fail("The stop of a stopped probe should fail with ProbeNotStartedException");
        } catch (final ProbeStopException e) {
            TestCase.fail(e.getMessage());
        } catch (final ProbeNotInitializedException e) {
            TestCase.fail(e.getMessage());
        } catch (final ProbeNotStartedException e) {
            // NOP, its the expected exception
            TestCase.assertTrue("The probe is no more initialized", testProbe.isInitialized);
            TestCase.assertFalse("The probe is started", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific stop method 'doStop' should not be invoked on a stop of an stopped probe.",
                    1, testProbe.specificStopDoneCount);
        }
    }

    /**
     * <p>
     * Tests the stop of a shutdown probe.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the exception {@link ProbeNotInitializedException} is thrown,</li>
     * <li>the specific method 'doStop' is not invoked,</li>
     * <li>the probe status does not change, always 'not initialized'.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_109() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization, startup, stop and shutdown, no exception is expected.
        try {
            testProbe.init();
            testProbe.start();
            testProbe.stop();
            testProbe.shutdown();

            TestCase.assertFalse("The probe is initialized", testProbe.isInitialized);
            TestCase.assertFalse("The probe is started", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' was not invoked once.", 1,
                    testProbe.specificInitializationDoneCount);
            TestCase.assertEquals("The specific startup method 'doStart' was not invoked once.", 1,
                    testProbe.specificStartupDoneCount);
            TestCase.assertEquals("The specific stop method 'doStop' was not invoked once.", 1,
                    testProbe.specificStopDoneCount);
            TestCase.assertEquals(
                    "The specific shutdown method 'doShutdown' was not invoked once.", 1,
                    testProbe.specificShutdownDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }

        // Stop without startup, ProbeNotStartedException is
        // expected
        try {
            testProbe.stop();
            TestCase.fail("The stop of a shutdown probe should fail with ProbeNotInitializedException");
        } catch (final ProbeStopException e) {
            TestCase.fail(e.getMessage());
        } catch (final ProbeNotInitializedException e) {
            // NOP, its the expected exception
            TestCase.assertFalse("The probe was re-initialized", testProbe.isInitialized);
            TestCase.assertFalse("The probe was re-started", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific stop method 'doStop' should not be invoked on a stop of a shutdown probe.",
                    1, testProbe.specificStopDoneCount);
        } catch (final ProbeNotStartedException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the shutdown of a just allocated probe.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the exception {@link ProbeNotInitializedException} is thrown,</li>
     * <li>the specific method 'doShutdown' is not invoked,</li>
     * <li>the probe status does not change, always 'not initialized'.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_110() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Shutdown without initialization, ProbeNotInitiliazedException is
        // expected
        try {
            testProbe.shutdown();
            TestCase.fail("The shutdown of a not initialized probe should fail with ProbeNotInitializedException");
        } catch (final ProbeNotInitializedException e) {
            // NOP, its the expected exception
            TestCase.assertFalse("The probe should not be initialized", testProbe.isInitialized);
            TestCase.assertFalse("The probe should not be started", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific shutdown method 'doShutdown' should not be invoked on a shutdown of a just allocated probe.",
                    0, testProbe.specificShutdownDoneCount);
        } catch (final ProbeShutdownException e) {
            TestCase.fail(e.getMessage());
        } catch (final ProbeStartedException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the shutdown of a started probe.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the exception {@link ProbeStartedException} is thrown,</li>
     * <li>the specific method 'doShutdown' is not invoked,</li>
     * <li>the probe status does not change, always 'started'.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_111() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization and startup, no exception is expected.
        try {
            testProbe.init();
            testProbe.start();

            TestCase.assertTrue("The probe is not initialized", testProbe.isInitialized);
            TestCase.assertTrue("The probe is not started", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' was not invoked once.", 1,
                    testProbe.specificInitializationDoneCount);
            TestCase.assertEquals("The specific startup method 'doStart' was not invoked once.", 1,
                    testProbe.specificStartupDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }

        // Shutdown without stop, ProbeStartedException is
        // expected
        try {
            testProbe.shutdown();
            TestCase.fail("The shutdown of a started probe should fail with ProbeStartedException");
        } catch (final ProbeShutdownException e) {
            TestCase.fail(e.getMessage());
        } catch (final ProbeNotInitializedException e) {
            TestCase.fail(e.getMessage());
        } catch (final ProbeStartedException e) {
            // NOP, its the expected exception
            TestCase.assertTrue("The probe should be initialized", testProbe.isInitialized);
            TestCase.assertTrue("The probe should be started", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific shutdown method 'doShutdown' should not be invoked on a shutdown of a started probe.",
                    0, testProbe.specificShutdownDoneCount);
        }
    }

    /**
     * <p>
     * Tests the shutdown of a shutdown probe.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the exception {@link ProbeNotInitializedException} is thrown,</li>
     * <li>the specific method 'doShutdown' is not invoked,</li>
     * <li>the probe status does not change, always 'not initialized'.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_112() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, false);

        // Initialization, startup, stop and shutdown, no exception is expected.
        try {
            testProbe.init();
            testProbe.start();
            testProbe.stop();
            testProbe.shutdown();

            TestCase.assertFalse("The probe is initialized", testProbe.isInitialized);
            TestCase.assertFalse("The probe is started", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific initialization method 'doInit' was not invoked once.", 1,
                    testProbe.specificInitializationDoneCount);
            TestCase.assertEquals("The specific startup method 'doStart' was not invoked once.", 1,
                    testProbe.specificStartupDoneCount);
            TestCase.assertEquals("The specific stop method 'doStop' was not invoked once.", 1,
                    testProbe.specificStopDoneCount);
            TestCase.assertEquals(
                    "The specific shutdown method 'doShutdown' was not invoked once.", 1,
                    testProbe.specificShutdownDoneCount);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }

        // Shutdown, ProbeNotInitializedException is expected
        try {
            testProbe.shutdown();
            TestCase.fail("The shutdown of a shutdown probe should fail with ProbeNotInitializedException");
        } catch (final ProbeNotInitializedException e) {
            // NOP, its the expected exception
            TestCase.assertFalse("The probe should not be initialized", testProbe.isInitialized);
            TestCase.assertFalse("The probe should not be started", testProbe.isStarted);
            TestCase.assertEquals(
                    "The specific shutdown method 'doShutdown' should not be invoked on a shutdown of a shutdown probe.",
                    1, testProbe.specificShutdownDoneCount);
        } catch (final ProbeShutdownException e) {
            TestCase.fail(e.getMessage());
        } catch (final ProbeStartedException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests a specific initialization with error.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the exception {@link ProbeInitializationException} is thrown,</li>
     * <li>the probe isn't initialized.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_113() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(true, false,
                false, false);

        // Initialization
        try {
            testProbe.init();
            TestCase.fail("The exception ProbeInitializationException should be thrown.");
        } catch (final ProbeInitializationException e) {
            // It's the expected exception
            TestCase.assertFalse(
                    "The probe should not be initialized if the exception ProbeInitializationException is thrown during specific initialization.",
                    testProbe.isInitialized);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests a specific startup with error.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the exception {@link ProbeStartupException} is thrown,</li>
     * <li>the probe isn't started.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_114() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, true,
                false, false);

        // Initialization and startup
        try {
            testProbe.init();
            testProbe.start();
            TestCase.fail("The exception ProbeStartupException should be thrown.");
        } catch (final ProbeStartupException e) {
            // It's the expected exception
            TestCase.assertFalse(
                    "The probe should not be started if the exception ProbeStartupeException is thrown during specific startup.",
                    testProbe.isStarted);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests a specific stop with error.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the exception {@link ProbeStopException} is thrown,</li>
     * <li>the probe is stopped.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_115() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                true, false);

        // Initialization, startup and stop
        try {
            testProbe.init();
            testProbe.start();
            testProbe.stop();
            TestCase.fail("The exception ProbeStopException should be thrown.");
        } catch (final ProbeStopException e) {
            // It's the expected exception
            TestCase.assertFalse(
                    "The probe should be stopped (status = not started) if the exception ProbeStopException is thrown during specific stop.",
                    testProbe.isStarted);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests a specific shutdown with error.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the exception {@link ProbeShutdownException} is thrown,</li>
     * <li>the probe is shutdown.</li>
     * </ul>
     * </p>
     */
    public void testLifecycle_116() {
        final TestProbeSpecificDoMethods testProbe = new TestProbeSpecificDoMethods(false, false,
                false, true);

        // Initialization and shutdown
        try {
            testProbe.init();
            testProbe.shutdown();
            TestCase.fail("The exception ProbeShutdownException should be thrown.");
        } catch (final ProbeShutdownException e) {
            // It's the expected exception
            TestCase.assertFalse(
                    "The probe should be shutdown (status = not initialized) if the exception ProbeShutdownException is thrown during specific shutdown.",
                    testProbe.isInitialized);
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }
}
