/**
 * Copyright (c) 2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;

import junit.framework.TestCase;

import org.junit.Test;
import org.ow2.petals.probes.api.exceptions.ProbeException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;
import org.ow2.petals.probes.api.probes.GaugeProbe;
import org.ow2.petals.probes.api.sensor.GaugeSensor;
import org.ow2.petals.probes.api.sensor.detector.GaugeDefectDetector;

/**
 * <p>
 * Unit tests of {@link GaugeProbe} to verify:
 * <ul>
 * <li>access to instant value and max value</li>
 * <li>the thread-safety</li>
 * <li>specific methods against the lifecycle.</li>
 * </ul>
 * </p>
 * 
 * @author Christophe DENEUX (PetalsLink)
 * 
 */
public class GaugeProbeImplTestCase {

    private final static short THREAD_POOL_SIZE = 100;

    private final static long THREADSAFETY_ITERATION_NB = 10000;

    /**
     * A sensor incrementing its instant value.
     * 
     * @author Christophe DENEUX (PetalsLink)
     * 
     */
    private class PickCounterSensor implements GaugeSensor<Long, Long> {

        private AtomicLong counter;

        @Override
        public Long getInitialValue() {
            this.counter = new AtomicLong(0);
            return 0L;
        }

        @Override
        public Long getInstantValue() {
            return this.counter.incrementAndGet();
        }

        @Override
        public Long toExternalValue(final Long valueToConvert) {
            return valueToConvert;
        }
    };

    /**
     * A {@link GaugeDefectDetector} associated to the sensor
     * {@link PickCounterSensor}
     * 
     * @author Christophe DENEUX (PetalsLink)
     * 
     */
    private class PickCounterDefectDetector implements GaugeDefectDetector<Long> {

        final AtomicInteger maxValueReached;

        public PickCounterDefectDetector(final AtomicInteger maxValueReached) {
            this.maxValueReached = maxValueReached;
        }

        @Override
        public void detect(final Long instantValue) {
            if (instantValue == 2) {
                // The max value is reached
                maxValueReached.incrementAndGet();
            }
        }
    }

    /**
     * A {@link GaugeDefectDetector} that detects nothing ;-)
     * 
     * @author Christophe DENEUX (PetalsLink)
     * 
     */
    private class DummyDefectDetector implements GaugeDefectDetector<Long> {

        @Override
        public void detect(Long instantValue) {
            // NOP
        }

    }

    /**
     * <p>
     * Tests the method 'pick' of a started probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>the max value has the right value after a 'pick',</li>
     * <li>a defect is detected when the max value is reach.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testPick_001() {
        final AtomicInteger maxValueReached = new AtomicInteger(0);
        final GaugeProbe<Long, Long> gaugeProbe = new GaugeProbeImpl<Long, Long>(
                new PickCounterSensor(), new PickCounterDefectDetector(maxValueReached));

        try {
            gaugeProbe.init();
            gaugeProbe.start();
            TestCase.assertEquals("invalid max value just after startup.", new Long(0),
                    gaugeProbe.getMaxValue());
            gaugeProbe.pick();
            TestCase.assertEquals("invalid max value.", Long.valueOf(1), gaugeProbe.getMaxValue());
            gaugeProbe.pick();
            TestCase.assertEquals("invalid max value.", Long.valueOf(2), gaugeProbe.getMaxValue());
            TestCase.assertEquals("error in the defect detection", 1, maxValueReached.intValue());

        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the method 'pick' of a re-started probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>the max value has the right value after a 'pick' occurring after the
     * first startup'.</li>
     * <li>the max value is reset when restarting the probe,</li>
     * <li>the max value has the right value after a 'pick' occurring after the
     * 2nd startup',</li>
     * <li>a defect is detected when the max value is reach.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testPick_002() {
        final AtomicInteger maxValueReached = new AtomicInteger(0);
        final GaugeProbe<Long, Long> gaugeProbe = new GaugeProbeImpl<Long, Long>(
                new PickCounterSensor(), new PickCounterDefectDetector(maxValueReached));

        try {
            gaugeProbe.init();
            gaugeProbe.start();
            TestCase.assertEquals("invalid max value just after startup.", Long.valueOf(0),
                    gaugeProbe.getMaxValue());
            gaugeProbe.pick();
            TestCase.assertEquals("invalid max value.", Long.valueOf(1), gaugeProbe.getMaxValue());
            gaugeProbe.pick();
            TestCase.assertEquals("invalid max value.", Long.valueOf(2), gaugeProbe.getMaxValue());
            TestCase.assertEquals("error in the defect detection", 1, maxValueReached.intValue());
            gaugeProbe.stop();
            gaugeProbe.start();
            TestCase.assertEquals("invalid max value just after restart.", Long.valueOf(0),
                    gaugeProbe.getMaxValue());
            gaugeProbe.pick();
            TestCase.assertEquals("invalid max value.", Long.valueOf(1), gaugeProbe.getMaxValue());

        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the method 'getInstantValue' of a just started probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>the instant value is returned.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetInstantValue_001() {
        final AtomicInteger maxValueReached = new AtomicInteger(0);
        final GaugeProbe<Long, Long> gaugeProbe = new GaugeProbeImpl<Long, Long>(
                new PickCounterSensor(), new PickCounterDefectDetector(maxValueReached));

        try {
            gaugeProbe.init();
            gaugeProbe.start();
            TestCase.assertEquals("The instant value is not the expected one.", Long.valueOf(1),
                    gaugeProbe.getInstantValue());
            TestCase.assertEquals("The instant value is not the expected one.", Long.valueOf(2),
                    gaugeProbe.getInstantValue());
            TestCase.assertEquals("error in the defect detection", 1, maxValueReached.intValue());

        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the method 'getMaxValue' of an initialized probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>just after the initialization, the max value has its initial value,</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetMaxValue_001() {
        final GaugeProbe<Long, Long> gaugeProbe = new GaugeProbeImpl<Long, Long>(
                new PickCounterSensor(), null);

        try {
            gaugeProbe.init();

            TestCase.assertEquals("The max value has not its initial value.", Long.valueOf(0),
                    gaugeProbe.getMaxValue());

        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the method 'getMaxValue' of a started probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>just after the initialization, the max value has its initial value,</li>
     * <li>several 'pick' after the startup, the max value has the right value.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetMaxValue_002() {
        final GaugeProbe<Long, Long> gaugeProbe = new GaugeProbeImpl<Long, Long>(
                new PickCounterSensor(), null);

        try {
            gaugeProbe.init();
            TestCase.assertEquals("The max value has not its initial value.", Long.valueOf(0),
                    gaugeProbe.getMaxValue());

            gaugeProbe.start();
            gaugeProbe.pick();
            gaugeProbe.pick();
            TestCase.assertEquals("The max value has not the expected value.", Long.valueOf(2),
                    gaugeProbe.getMaxValue());

        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the method 'getMaxValue' of a stopped probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>just after the initialization, the max value has its initial value,</li>
     * <li>several 'pick' after the startup, the max value has the right value.</li>
     * <li>after to be stop, the max value has as the same value as before the
     * stop.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetMaxValue_003() {
        final GaugeProbe<Long, Long> gaugeProbe = new GaugeProbeImpl<Long, Long>(
                new PickCounterSensor(), null);

        try {
            gaugeProbe.init();
            TestCase.assertEquals("The max value has not its initial value.", Long.valueOf(0),
                    gaugeProbe.getMaxValue());

            gaugeProbe.start();
            gaugeProbe.pick();
            gaugeProbe.pick();
            TestCase.assertEquals("The max value has not the expected value.", Long.valueOf(2),
                    gaugeProbe.getMaxValue());

            gaugeProbe.stop();
            TestCase.assertEquals("The max value has not the expected value after the stop.",
                    Long.valueOf(2), gaugeProbe.getMaxValue());

        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /*----------------------------*/
    /* Error test cases */
    /*----------------------------*/

    /**
     * <p>
     * Tests the constructor against a sensor set to <code>null</code>.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link IllegalArgumentException} is thrown,</li>
     * </ul>
     * </p>
     */
    @Test
    public void testConstructor_101() {
        try {
            new GaugeProbeImpl<Long, Long>(null, new DummyDefectDetector());
            TestCase.fail("The constructor must throw IllegalArgumentException when the sensor is null.");
        } catch (final IllegalArgumentException e) {
            // NOP, it's the expected exception
        }
    }

    /**
     * <p>
     * Tests the method 'pick' of a just allocated probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * </ul>
     * </p>
     */
    @Test
    public void testPick_101() {
        final GaugeProbe<Long, Long> gaugeProbe = new GaugeProbeImpl<Long, Long>(
                new PickCounterSensor(), null);

        try {
            gaugeProbe.pick();
            TestCase.fail("The 'pick' of a just allocated probe should fail with ProbeNotStartedException");

        } catch (final ProbeNotStartedException e) {
            // NOP, it's the expected exception
        }
    }

    /**
     * <p>
     * Tests the method 'pick' of a stopped probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>after to be stop, the max value has as the same value as before the
     * stop.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testPick_102() {
        final GaugeProbe<Long, Long> gaugeProbe = new GaugeProbeImpl<Long, Long>(
                new PickCounterSensor(), null);

        try {
            gaugeProbe.init();
            gaugeProbe.start();
            gaugeProbe.pick();
            gaugeProbe.stop();
            gaugeProbe.pick();
            TestCase.fail("The 'pick' of a stopped probe should fail with ProbeNotStartedException");

        } catch (final ProbeNotStartedException e) {
            // NOP, it's the expected exception
            try {
                TestCase.assertEquals("The max value has not the expected value after the stop.",
                        Long.valueOf(1), gaugeProbe.getMaxValue());
            } catch (ProbeNotInitializedException e1) {
                TestCase.fail(e.getMessage());
            }
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the method 'pick' of a shutdown probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * </ul>
     * </p>
     */
    @Test
    public void testPick_103() {
        final GaugeProbe<Long, Long> gaugeProbe = new GaugeProbeImpl<Long, Long>(
                new PickCounterSensor(), null);

        try {
            gaugeProbe.init();
            gaugeProbe.start();
            gaugeProbe.pick();
            gaugeProbe.stop();
            gaugeProbe.shutdown();
            gaugeProbe.pick();
            TestCase.fail("The 'pick' of a shutdown probe should fail with ProbeNotStartedException");

        } catch (final ProbeNotStartedException e) {
            // NOP, it's the expected exception
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the method 'getInstantValue' of a just allocated probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetInstantValue_101() {
        final GaugeProbe<Long, Long> gaugeProbe = new GaugeProbeImpl<Long, Long>(
                new PickCounterSensor(), null);

        try {
            gaugeProbe.getInstantValue();
            TestCase.fail("The methods 'getInstantValue' of a just allocated probe should fail with ProbeNotStartedException");

        } catch (final ProbeNotStartedException e) {
            // NOP, it's the expected exception
        }
    }

    /**
     * <p>
     * Tests the method 'getInstantValue' of a stopped probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetInstantValue_102() {
        final GaugeProbe<Long, Long> gaugeProbe = new GaugeProbeImpl<Long, Long>(
                new PickCounterSensor(), null);

        try {
            gaugeProbe.init();
            gaugeProbe.start();
            gaugeProbe.stop();
            gaugeProbe.getInstantValue();
            TestCase.fail("The method 'getInstantValue' of a stopped probe should fail with ProbeNotStartedException");

        } catch (final ProbeNotStartedException e) {
            // NOP, it's the expected exception
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the method 'getInstantValue' of a shutdown probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetInstantValue_103() {
        final GaugeProbe<Long, Long> gaugeProbe = new GaugeProbeImpl<Long, Long>(
                new PickCounterSensor(), null);

        try {
            gaugeProbe.init();
            gaugeProbe.start();
            gaugeProbe.stop();
            gaugeProbe.shutdown();
            gaugeProbe.getInstantValue();
            TestCase.fail("The method 'getInstantValue' of a shutdown probe should fail with ProbeNotStartedException");

        } catch (final ProbeNotStartedException e) {
            // NOP, it's the expected exception
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the method 'getMaxValue' of a just allocated probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotInitializedException} is thrown,</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetMaxValue_101() {
        final GaugeProbe<Long, Long> gaugeProbe = new GaugeProbeImpl<Long, Long>(
                new PickCounterSensor(), null);

        try {
            gaugeProbe.getMaxValue();
            TestCase.fail("The methods 'getMaxValue' of a just allocated probe should fail with ProbeNotInitializedException");

        } catch (final ProbeNotInitializedException e) {
            // NOP, it's the expected exception
        }
    }

    /**
     * <p>
     * Tests the method 'getMaxValue' of a shutdown probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotInitializedException} is thrown,</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetMaxValue_102() {
        final GaugeProbe<Long, Long> gaugeProbe = new GaugeProbeImpl<Long, Long>(
                new PickCounterSensor(), null);

        try {
            gaugeProbe.init();
            gaugeProbe.start();
            gaugeProbe.stop();
            gaugeProbe.shutdown();
            gaugeProbe.getMaxValue();
            TestCase.fail("The method 'getMaxValue' of a stopped probe should fail with ProbeNotInitializedException");

        } catch (final ProbeNotInitializedException e) {
            // NOP, it's the expected exception
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /*--------------------------*/
    /* Thread-safety test cases */
    /*--------------------------*/

    /**
     * <p>
     * Tests the thread-safety of the methods 'pick', 'getInstantValue' and
     * 'getMaxValue'. No change is made on the probe lifecycle.
     * </p>
     * 
     */
    @Test
    public void testThreadSafety_001() throws ProbeException, InterruptedException {

        // Testcase failures
        final List<Throwable> testcaseFailures = Collections
                .synchronizedList(new LinkedList<Throwable>());

        // Synchronization objects
        final Object starterMonitor = new Object();
        final Semaphore readyMonitor = new Semaphore(THREAD_POOL_SIZE);

        // Initialize and start the probe
        final GaugeProbe<Long, Long> gaugeProbe = new GaugeProbeImpl<Long, Long>(
                new PickCounterSensor(), null);
        gaugeProbe.init();
        gaugeProbe.start();

        // Initialize and start the thread pool size to THREAD_POOL_SIZE
        final List<Thread> threads = new LinkedList<Thread>();
        for (int i = 0; i < THREAD_POOL_SIZE; i++) {
            readyMonitor.acquire();

            final Thread thread = new Thread(new Runnable() {

                @Override
                public void run() {

                    try {
                        synchronized (starterMonitor) {
                            readyMonitor.release();
                            starterMonitor.wait();
                        }

                        for (int j = 0; j < THREADSAFETY_ITERATION_NB; j++) {
                            gaugeProbe.pick();
                            gaugeProbe.getInstantValue();
                            gaugeProbe.getMaxValue();
                        }
                    } catch (final Throwable e) {
                        testcaseFailures.add(e);
                    }

                }
            });
            threads.add(thread);
            thread.start();
        }

        // Wait that threads end to start
        readyMonitor.acquire(THREAD_POOL_SIZE);

        synchronized (starterMonitor) {
            starterMonitor.notifyAll();
        }

        // Wait the end of processing
        for (Thread thread : threads) {
            thread.join();
        }

        TestCase.assertEquals("Error occurs: " + testcaseFailures.toString(), 0,
                testcaseFailures.size());
        TestCase.assertEquals("The expected max value is not the expected one",
                Long.valueOf(THREAD_POOL_SIZE * THREADSAFETY_ITERATION_NB * 2),
                gaugeProbe.getMaxValue());

    }

    /**
     * Check that no error occurs on start/stop cycles (see PETALSCDK-116)
     */
    @Test
    public void startStopCycles() throws ProbeInitializedException, ProbeStartedException,
            ProbeInitializationException, ProbeNotInitializedException, ProbeNotStartedException, ProbeStopException,
            ProbeStartupException, ProbeShutdownException {

        final GaugeProbe<Long, Long> gaugeProbe = new GaugeProbeImpl<Long, Long>(new PickCounterSensor(), null);
        gaugeProbe.init();

        gaugeProbe.start();
        gaugeProbe.stop();

        gaugeProbe.start();
        gaugeProbe.stop();

        gaugeProbe.shutdown();
    }
    
    // TODO: Add a test about thread-safety of pick, getInstantValue and lifecycle changes
}
