/**
 * Copyright (c) 2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicLong;

import junit.framework.TestCase;

import org.junit.Test;
import org.ow2.petals.probes.api.exceptions.ProbeException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;
import org.ow2.petals.probes.api.probes.KeyedGaugeProbe;
import org.ow2.petals.probes.api.sensor.KeyedGaugeSensor;
import org.ow2.petals.probes.key.ProbeTestKey;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

/**
 * <p>
 * Unit tests of {@link KeyedGaugeProbe} to verify:
 * <ul>
 * <li>access to instant value and max value by key</li>
 * <li>the thread-safety</li>
 * <li>specific methods against the lifecycle.</li>
 * </ul>
 * </p>
 * 
 * @author Christophe DENEUX (PetalsLink)
 * 
 */
public class KeyedGaugeProbeImplTestCase {

	private final static short THREAD_POOL_SIZE = 25;

	private final static long THREADSAFETY_ITERATION_NB = 10000;

	private final static short KEYS_NUMBER = 4;

	/**
	 * A sensor incrementing its instant value.
	 * 
	 * @author Christophe DENEUX (PetalsLink)
	 * 
	 */
	private class PickCounterSensor implements
			KeyedGaugeSensor<ProbeTestKey, Long, Long> {

		private ConcurrentMap<ProbeTestKey, AtomicLong> counters = new ConcurrentHashMap<ProbeTestKey, AtomicLong>(0);

		@Override
		public Long toExternalValue(final Long valueToConvert) {
			return valueToConvert;
		}

		@Override
		public Long getInstantValue(final ProbeTestKey key) {
			final AtomicLong initialValueIfNeeded = new AtomicLong(1);
			final AtomicLong value = this.counters.putIfAbsent(key,
					initialValueIfNeeded);
			if (value != null) {
				return value.incrementAndGet();
			}
			else {
				return Long.valueOf(1);
			}
		}

		@Override
		public void reset() {
			this.counters.clear();
			
		}
	};

	/**
	 * <p>
	 * Tests the method 'pick' of a started probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>no exception is thrown,</li>
	 * <li>the max values have the right values after a 'pick'.</li>
	 * </ul>
	 * </p>
	 */
    @Test
	public void testPick_001() {
        final KeyedGaugeProbe<ProbeTestKey, Long, Long> gaugeProbe = new KeyedGaugeProbeImpl<ProbeTestKey, Long, Long>(
				new PickCounterSensor());
		final ProbeTestKey key1 = new ProbeTestKey("key1");
		final ProbeTestKey key2 = new ProbeTestKey("key2");

		try {
			gaugeProbe.init();
			gaugeProbe.start();
			TestCase.assertEquals(
					"At least one key is available in the probe.", 0,
					gaugeProbe.getMaxValues().size());
			gaugeProbe.pick(key1);
			gaugeProbe.pick(key2);
			TestCase.assertEquals("invalid max value for key1.",
					Long.valueOf(1), gaugeProbe.getMaxValues().get(key1));
			TestCase.assertEquals("invalid max value for key2.",
					Long.valueOf(1), gaugeProbe.getMaxValues().get(key2));
			gaugeProbe.pick(key1);
			TestCase.assertEquals("invalid max value for key1.",
					Long.valueOf(2), gaugeProbe.getMaxValues().get(key1));
			TestCase.assertEquals("invalid max value for key2.",
					Long.valueOf(1), gaugeProbe.getMaxValues().get(key2));

		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'pick' of a re-started probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>no exception is thrown,</li>
	 * <li>the max values haves the right values after a 'pick' occurring after
	 * the first startup'.</li>
	 * <li>the max values are reset when restarting the probe,</li>
	 * <li>the max values have the right values after a 'pick' occurring after
	 * the 2nd startup'.</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testPick_002() {
        final KeyedGaugeProbe<ProbeTestKey, Long, Long> gaugeProbe = new KeyedGaugeProbeImpl<ProbeTestKey, Long, Long>(
				new PickCounterSensor());
		final ProbeTestKey key1 = new ProbeTestKey("key1");
		final ProbeTestKey key2 = new ProbeTestKey("key2");

		try {
			gaugeProbe.init();
			gaugeProbe.start();
			TestCase.assertEquals(
					"At least one key is available in the probe.", 0,
					gaugeProbe.getMaxValues().size());
			gaugeProbe.pick(key1);
			gaugeProbe.pick(key2);
			TestCase.assertEquals("invalid max value for key1.",
					Long.valueOf(1), gaugeProbe.getMaxValues().get(key1));
			TestCase.assertEquals("invalid max value for key2.",
					Long.valueOf(1), gaugeProbe.getMaxValues().get(key2));
			gaugeProbe.pick(key1);
			TestCase.assertEquals("invalid max value for key1.",
					Long.valueOf(2), gaugeProbe.getMaxValues().get(key1));
			TestCase.assertEquals("invalid max value for key2.",
					Long.valueOf(1), gaugeProbe.getMaxValues().get(key2));
			gaugeProbe.stop();
			gaugeProbe.start();
			TestCase.assertEquals(
					"At least one key is available in the probe.", 0,
					gaugeProbe.getMaxValues().size());
			gaugeProbe.pick(key1);
			TestCase.assertEquals("invalid max value for key1.",
					Long.valueOf(1), gaugeProbe.getMaxValues().get(key1));
			TestCase.assertEquals(
					"At least two keys are available in the probe.", 1,
					gaugeProbe.getMaxValues().size());

		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'getInstantValue' of a just started probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>no exception is thrown,</li>
	 * <li>the instant values are returned accordingly previous encountered
	 * keys.</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testGetInstantValues_001() {
        final KeyedGaugeProbe<ProbeTestKey, Long, Long> gaugeProbe = new KeyedGaugeProbeImpl<ProbeTestKey, Long, Long>(
				new PickCounterSensor());
		final ProbeTestKey key1 = new ProbeTestKey("key1");
		final ProbeTestKey key2 = new ProbeTestKey("key2");

		try {
			gaugeProbe.init();
			gaugeProbe.start();
			TestCase.assertEquals(
					"At least one key is available in the probe.", 0,
					gaugeProbe.getMaxValues().size());
			gaugeProbe.pick(key1);
			TestCase.assertEquals("More than one instant values is available.",
					1, gaugeProbe.getInstantValues().size());
			TestCase.assertNotNull("The instant value of 'key1' is missing.",
					gaugeProbe.getInstantValues().get(key1));
			TestCase.assertEquals("The instant value of 'key1' is incorrect.",
					Long.valueOf(4), gaugeProbe.getInstantValues().get(key1));
			TestCase.assertNull("The instant value of 'key2' is present.",
					gaugeProbe.getInstantValues().get(key2));

		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'getMaxValue' of an initialized probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>no exception is thrown,</li>
	 * <li>no max value is available,</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testGetMaxValues_001() {
        final KeyedGaugeProbe<ProbeTestKey, Long, Long> gaugeProbe = new KeyedGaugeProbeImpl<ProbeTestKey, Long, Long>(
				new PickCounterSensor());

		try {
			gaugeProbe.init();

			TestCase.assertEquals(
					"At least one max value is available in the probe.", 0,
					gaugeProbe.getMaxValues().size());

		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'getMaxValue' of a started probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>no exception is thrown,</li>
	 * <li>just after the initialization, no max value is available,</li>
	 * <li>several 'pick' after the startup, the max value has the right value.</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testGetMaxValues_002() {
        final KeyedGaugeProbe<ProbeTestKey, Long, Long> gaugeProbe = new KeyedGaugeProbeImpl<ProbeTestKey, Long, Long>(
				new PickCounterSensor());
		final ProbeTestKey key1 = new ProbeTestKey("key1");
		final ProbeTestKey key2 = new ProbeTestKey("key2");

		try {
			gaugeProbe.init();
			TestCase.assertEquals(
					"At least one max value is available in the probe.", 0,
					gaugeProbe.getMaxValues().size());

			gaugeProbe.start();
			gaugeProbe.pick(key1);
			gaugeProbe.pick(key1);
			gaugeProbe.pick(key2);
			TestCase.assertEquals(
					"The max value of 'key1' has not the expected value.",
					Long.valueOf(2), gaugeProbe.getMaxValues().get(key1));
			TestCase.assertEquals(
					"The max value of 'key2' has not the expected value.",
					Long.valueOf(1), gaugeProbe.getMaxValues().get(key2));

		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'getMaxValue' of a stopped probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>no exception is thrown,</li>
	 * <li>just after the initialization, no max value is available,</li>
	 * <li>several 'pick' after the startup, the max value has the right value.</li>
	 * <li>after to be stop, the max value has as the same value as before the
	 * stop.</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testGetMaxValues_003() {
        final KeyedGaugeProbe<ProbeTestKey, Long, Long> gaugeProbe = new KeyedGaugeProbeImpl<ProbeTestKey, Long, Long>(
				new PickCounterSensor());
		final ProbeTestKey key1 = new ProbeTestKey("key1");
		final ProbeTestKey key2 = new ProbeTestKey("key2");

		try {
			gaugeProbe.init();
			TestCase.assertEquals(
					"At least one max value is available in the probe.", 0,
					gaugeProbe.getMaxValues().size());

			gaugeProbe.start();
			gaugeProbe.pick(key1);
			gaugeProbe.pick(key1);
			gaugeProbe.pick(key2);
			TestCase.assertEquals(
					"The max value of 'key1' has not the expected value.",
					Long.valueOf(2), gaugeProbe.getMaxValues().get(key1));
			TestCase.assertEquals(
					"The max value of 'key2' has not the expected value.",
					Long.valueOf(1), gaugeProbe.getMaxValues().get(key2));

			gaugeProbe.stop();
			TestCase.assertEquals(
					"The max value of 'key1' has not the expected value.",
					Long.valueOf(2), gaugeProbe.getMaxValues().get(key1));
			TestCase.assertEquals(
					"The max value of 'key2' has not the expected value.",
					Long.valueOf(1), gaugeProbe.getMaxValues().get(key2));

		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/*----------------------------*/
	/* Lifecycle error test cases */
	/*----------------------------*/

	/**
	 * <p>
	 * Tests the method 'pick' of a just allocated probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testPick_101() {
        final KeyedGaugeProbe<ProbeTestKey, Long, Long> gaugeProbe = new KeyedGaugeProbeImpl<ProbeTestKey, Long, Long>(
				new PickCounterSensor());
		final ProbeTestKey key1 = new ProbeTestKey("key1");

		try {
			gaugeProbe.pick(key1);
			TestCase.fail("The 'pick' of a just allocated probe should fail with ProbeNotStartedException");

		} catch (final ProbeNotStartedException e) {
			// NOP, it's the expected exception
		}
	}

	/**
	 * <p>
	 * Tests the method 'pick' of a stopped probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
	 * <li>after to be stop, the max value has as the same value as before the
	 * stop.</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testPick_102() {
        final KeyedGaugeProbe<ProbeTestKey, Long, Long> gaugeProbe = new KeyedGaugeProbeImpl<ProbeTestKey, Long, Long>(
				new PickCounterSensor());
		final ProbeTestKey key1 = new ProbeTestKey("key1");

		try {
			gaugeProbe.init();
			gaugeProbe.start();
			gaugeProbe.pick(key1);
			gaugeProbe.stop();
			gaugeProbe.pick(key1);
			TestCase.fail("The 'pick' of a stopped probe should fail with ProbeNotStartedException");

		} catch (final ProbeNotStartedException e) {
			// NOP, it's the expected exception
			try {
				TestCase.assertEquals(
						"The max value has not the expected value after the stop.",
						Long.valueOf(1), gaugeProbe.getMaxValues().get(key1));
			} catch (ProbeNotInitializedException e1) {
				TestCase.fail(e.getMessage());
			}
		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'pick' of a shutdown probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testPick_103() {
        final KeyedGaugeProbe<ProbeTestKey, Long, Long> gaugeProbe = new KeyedGaugeProbeImpl<ProbeTestKey, Long, Long>(
				new PickCounterSensor());
		final ProbeTestKey key1 = new ProbeTestKey("key1");

		try {
			gaugeProbe.init();
			gaugeProbe.start();
			gaugeProbe.pick(key1);
			gaugeProbe.stop();
			gaugeProbe.shutdown();
			gaugeProbe.pick(key1);
			TestCase.fail("The 'pick' of a shutdown probe should fail with ProbeNotStartedException");

		} catch (final ProbeNotStartedException e) {
			// NOP, it's the expected exception
			TestCase.assertEquals(
					"At least one max value is available in the probe.", 0,
					((Map<?, Long>) ReflectionHelper.getPrivateFieldValue(KeyedGaugeProbeImpl.class, gaugeProbe, "maxValues")).size());
		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'getInstantValues' of a just allocated probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testGetInstantValues_101() {
        final KeyedGaugeProbe<ProbeTestKey, Long, Long> gaugeProbe = new KeyedGaugeProbeImpl<ProbeTestKey, Long, Long>(
				new PickCounterSensor());

		try {
			gaugeProbe.getInstantValues();
			TestCase.fail("The methods 'getInstantValues' of a just allocated probe should fail with ProbeNotStartedException");

		} catch (final ProbeNotStartedException e) {
			// NOP, it's the expected exception
		}
	}

	/**
	 * <p>
	 * Tests the method 'getInstantValues' of a stopped probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testGetInstantValues_102() {
        final KeyedGaugeProbe<ProbeTestKey, Long, Long> gaugeProbe = new KeyedGaugeProbeImpl<ProbeTestKey, Long, Long>(
				new PickCounterSensor());

		try {
			gaugeProbe.init();
			gaugeProbe.start();
			gaugeProbe.stop();
			gaugeProbe.getInstantValues();
			TestCase.fail("The method 'getInstantValues' of a stopped probe should fail with ProbeNotStartedException");

		} catch (final ProbeNotStartedException e) {
			// NOP, it's the expected exception
		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'getInstantValues' of a shutdown probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testGetInstantValues_103() {
        final KeyedGaugeProbe<ProbeTestKey, Long, Long> gaugeProbe = new KeyedGaugeProbeImpl<ProbeTestKey, Long, Long>(
				new PickCounterSensor());

		try {
			gaugeProbe.init();
			gaugeProbe.start();
			gaugeProbe.stop();
			gaugeProbe.shutdown();
			gaugeProbe.getInstantValues();
			TestCase.fail("The method 'getInstantValues' of a shutdown probe should fail with ProbeNotStartedException");

		} catch (final ProbeNotStartedException e) {
			// NOP, it's the expected exception
		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'getMaxValues' of a just allocated probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>the exception {@link ProbeNotInitializedException} is thrown,</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testGetMaxValues_101() {
        final KeyedGaugeProbe<ProbeTestKey, Long, Long> gaugeProbe = new KeyedGaugeProbeImpl<ProbeTestKey, Long, Long>(
				new PickCounterSensor());

		try {
			gaugeProbe.getMaxValues();
			TestCase.fail("The methods 'getMaxValues' of a just allocated probe should fail with ProbeNotInitializedException");

		} catch (final ProbeNotInitializedException e) {
			// NOP, it's the expected exception
		}
	}

	/**
	 * <p>
	 * Tests the method 'getMaxValues' of a shutdown probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>the exception {@link ProbeNotInitializedException} is thrown,</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testGetMaxValues_102() {
        final KeyedGaugeProbe<ProbeTestKey, Long, Long> gaugeProbe = new KeyedGaugeProbeImpl<ProbeTestKey, Long, Long>(
				new PickCounterSensor());

		try {
			gaugeProbe.init();
			gaugeProbe.start();
			gaugeProbe.stop();
			gaugeProbe.shutdown();
			gaugeProbe.getMaxValues();
			TestCase.fail("The method 'getMaxValues' of a stopped probe should fail with ProbeNotInitializedException");

		} catch (final ProbeNotInitializedException e) {
			// NOP, it's the expected exception
		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/*--------------------------*/
	/* Thread-safety test cases */
	/*--------------------------*/

	/**
	 * <p>
	 * Tests the thread-safety of the methods 'pick', 'getInstantValues' and
	 * 'getMaxValues'. No change is made on the probe lifecycle.
	 * </p>
	 * 
	 */
    @Test
    public void testThreadSafety_001() throws ProbeException,
			InterruptedException {

		// Testcase failures
		final List<Throwable> testcaseFailures = Collections
				.synchronizedList(new LinkedList<Throwable>());

		// Synchronization objects
		final Object starterMonitor = new Object();
		final Semaphore readyMonitor = new Semaphore(THREAD_POOL_SIZE);

		// Initialize and start the probe
        final KeyedGaugeProbe<ProbeTestKey, Long, Long> gaugeProbe = new KeyedGaugeProbeImpl<ProbeTestKey, Long, Long>(
				new PickCounterSensor());
		gaugeProbe.init();
		gaugeProbe.start();

		// Initialize and start the thread pool size to THREAD_POOL_SIZE
		final List<Thread> threads = new LinkedList<Thread>();
		for (int k = 0; k < KEYS_NUMBER; k++) {
			final ProbeTestKey key = new ProbeTestKey("key" + k);
			// Pick a value to initialize all keys
			gaugeProbe.pick(key);
			
			for (int i = 0; i < THREAD_POOL_SIZE; i++) {
				readyMonitor.acquire();

				final Thread thread = new Thread(new Runnable() {

					@Override
					public void run() {

						try {
							
							synchronized (starterMonitor) {
								readyMonitor.release();
								starterMonitor.wait();
							}

							for (int j = 0; j < THREADSAFETY_ITERATION_NB; j++) {
								gaugeProbe.pick(key);
								gaugeProbe.getInstantValues();
								gaugeProbe.getMaxValues();
							}
						} catch (final Throwable e) {
							testcaseFailures.add(e);
						}

					}
				});
				threads.add(thread);
				thread.start();
			}
		}

		// Wait that threads end to start
		readyMonitor.acquire(THREAD_POOL_SIZE);

		synchronized (starterMonitor) {
			starterMonitor.notifyAll();
		}

		// Wait the end of processing
		for (Thread thread : threads) {
			thread.join();
		}

		TestCase.assertEquals("Error occurs: " + testcaseFailures.toString(),
				0, testcaseFailures.size());

		for (int k = 0; k < KEYS_NUMBER; k++) {
			TestCase.assertEquals(
					"The expected max value is not the expected one",
					Long.valueOf(
							// Number of 'pick'
							THREAD_POOL_SIZE * THREADSAFETY_ITERATION_NB +
							// Number of 'getInstantvalues'
							THREAD_POOL_SIZE * THREADSAFETY_ITERATION_NB * KEYS_NUMBER +
							// Initial call to 'pick'
							1),
					gaugeProbe.getMaxValues().get(new ProbeTestKey("key" + k)));
		}
    }

    /**
     * Check that no error occurs on start/stop cycles (see PETALSCDK-116)
     */
    @Test
    public void startStopCycles() throws ProbeInitializedException, ProbeStartedException,
            ProbeInitializationException, ProbeNotInitializedException, ProbeNotStartedException, ProbeStopException,
            ProbeStartupException, ProbeShutdownException {

        final KeyedGaugeProbe<ProbeTestKey, Long, Long> gaugeProbe = new KeyedGaugeProbeImpl<ProbeTestKey, Long, Long>(
                new PickCounterSensor());
        gaugeProbe.init();

        gaugeProbe.start();
        gaugeProbe.stop();

        gaugeProbe.start();
        gaugeProbe.stop();

        gaugeProbe.shutdown();
    }

	// TODO: Add a test about thread-safety of pick, getInstantValue and
	// lifecycle changes
}
